"""
Read the license at:
https://github.com/IzunaDevs/nsfw_dl/blob/master/LICENSE
"""
import importlib
import io
import json
from urllib.parse import quote

import aiohttp
from bs4 import BeautifulSoup
import requests

from .errors import NoLoader, UnsupportedDataFormat

LOADERS = {
    "danbooru": ["DanbooruRandom", "DanbooruSearch"],
    "drunkenpumken": ["DrunkenpumkenRandom"],
    "e621": ["E621Random", "E621Search"],
    "furrybooru": ["FurrybooruRandom", "FurrybooruSearch"],
    "gelbooru": ["GelbooruRandom", "GelbooruSearch"],
    "hbrowse": ["HbrowseRandom"],
    "konachan": ["KonachanRandom", "KonachanSearch"],
    "lolibooru": ["LolibooruRandom", "LolibooruSearch"],
    "nhentai": ["NhentaiRandom"],
    "rule34": ["Rule34Random", "Rule34Search"],
    "tbib": ["TbibRandom", "TbibSearch"],
    "tsumino": ["TsuminoRandom"],
    "xbooru": ["XbooruRandom", "XbooruSearch"],
    "yandere": ["YandereRandom", "YandereSearch"],
}


class NSFWDL:
    """
    Main class.
    """
    def __init__(self, session=None, loop=None, json_loader=json.loads):
        self.async_ = None
        self.loop = loop
        self.session = session
        self.json_loader = json_loader
        self.loaders = {}
        self.load_default()

    def add_loader(self, name, downloader):
        """
        adds a loader to get images from.
        """
        self.loaders[name] = downloader

    @staticmethod
    def parse_args(args):
        """
        parses args.
        """
        return quote(args)

    async def __aenter__(self):
        if self.async_ is not True:
            if self.async_ is False:
                self.session.close()
            self.session = aiohttp.ClientSession(loop=self.loop)
        self.async_ = True
        return self

    async def __aexit__(self, *exc):
        return

    def __enter__(self):
        if self.async_ is not False:
            if self.async_ is True:
                self.session.close()

            self.session = requests.Session()
        self.async_ = False
        return self

    def __exit__(self, *exc):
        return

    def __del__(self):
        self.session.close()

    def __getattr__(self, item):
        if item in self.loaders:
            return self.loaders[item]

    async def get_async(self, url):
        async with self.session.get(url) as resp:
            return await resp.read()

    def get_sync(self, url):
        with self.session.get(url) as resp:
            return resp.content

    def get(self, url):
        return self.get_async(url) if self.async_ else self.get_sync(url)

    async def download_async(self, url, data, headers, loader, download=False):
        """ async downloader. """
        reqtype = getattr(loader, 'reqtype', None)
        if reqtype is None:
            reqtype = "get"
        reqmeth = getattr(self.session, reqtype)

        async with reqmeth(url, data=data, headers=headers) as resp:
            assert 200 <= resp.status < 300

            if loader.data_format == "bs4/html":
                reqdata = BeautifulSoup(await resp.text(), "html.parser")

            elif loader.data_format == "bs4/xml":
                reqdata = BeautifulSoup(await resp.text(), "lxml")

            elif loader.data_format == "json":
                reqdata = await resp.json(loads=self.json_loader)

            elif loader.data_format == "url":
                reqdata = loader.data_format

            else:
                raise UnsupportedDataFormat(loader.data_format)

        if reqdata == "url":
            img_url = resp.url

        else:
            img_url = loader.get_image(reqdata)

        if download:
            async with self.session.get(img_url) as resp:
                assert 200 <= resp.status < 300
                return io.BytesIO(await resp.read())

        return img_url

    def download_sync(self, url, data, headers, loader, download=False):
        """ sync downloader. """
        reqtype = getattr(loader, 'reqtype', None)
        if reqtype is None:
            reqtype = "get"
        reqmeth = getattr(self.session, reqtype)
        resp = reqmeth(url, data=data, headers=headers)
        assert 200 <= resp.status_code < 300

        if loader.data_format == "bs4/html":
            reqdata = BeautifulSoup(resp.text, "html.parser")

        elif loader.data_format == "bs4/xml":
            reqdata = BeautifulSoup(resp.text, "lxml")

        elif loader.data_format == "json":
            reqdata = self.json_loader(resp.text)

        elif loader.data_format == "url":
            reqdata = loader.data_format

        else:
            raise UnsupportedDataFormat(loader.data_format)

        if reqdata == "url":
            img_url = resp.url
        else:
            img_url = loader.get_image(reqdata)

        if download:
            resp = self.session.get(img_url)
            assert 200 <= resp.status < 300
            return io.BytesIO(resp.content)

        return img_url

    def download(self, name, args="", download=False):
        """
        downloads or returns the image urls based on the loaders.
        """
        if name not in self.loaders:
            raise NoLoader(f"No loader named {name!r}")

        loader = self.loaders[name]

        args = self.parse_args(args)
        url, data, headers = loader.prepare_url(args=args)

        return (self.download_async(url, data, headers, loader, download)
                if self.async_ else
                self.download_sync(url, data, headers, loader, download))

    def load_default(self):
        """
        loads the loaders.
        """
        for loader, names in LOADERS.items():
            lib = importlib.import_module(f"nsfw_dl.loaders.{loader}")
            for name, loader_class in [(name, getattr(lib, name))
                                       for name in names]:
                load_obj = loader_class()
                self.add_loader(name, load_obj)



"""
Read the license at:
https://github.com/IzunaDevs/nsfw_dl/blob/master/LICENSE
"""
from nsfw_dl.bases import BaseSearchHTML


class FurrybooruRandom:
    """ Gets a random image from furrybooru. """
    data_format = "bs4/html"

    @staticmethod
    def prepare_url(args):
        """ ... """
        type(args)
        return "http://furry.booru.org/index.php?page=post&s=random", {}, {}

    @staticmethod
    def get_image(data):
        """ ... """
        return data.find(id="image").get("src")


class FurrybooruSearch(BaseSearchHTML):
    """ Gets a random image with a specific tag from furrybooru. """
    data_format = "bs4/html"

    @staticmethod
    def prepare_url(args):
        """ ... """
        return ("http://furry.booru.org/index.php?page=post"
                f"&s=list&tags={args}", {}, {})



"""
Read the license at:
https://github.com/IzunaDevs/nsfw_dl/blob/master/LICENSE
"""
from nsfw_dl.bases import BaseSearchJSON


class YandereRandom:
    """ Gets a random image from yandere. """
    data_format = "bs4/html"

    @staticmethod
    def prepare_url(args):
        """ ... """
        type(args)
        return "https://yande.re/post/random", {}, {}

    @staticmethod
    def get_image(data):
        """ ... """
        return data.find(id="highres").get("href")


class YandereSearch(BaseSearchJSON):
    """ Gets a random image with a specific tag from yandere. """
    data_format = "json"

    @staticmethod
    def prepare_url(args):
        """ ... """
        return f"https://yande.re/post.json?tags={args}", {}, {}



"""
Read the license at:
https://github.com/IzunaDevs/nsfw_dl/blob/master/LICENSE
"""
import random

from nsfw_dl.errors import NoResultsFound

# base classes for some loaders.


class BaseSearchXML:
    """ base xml search class. """
    @staticmethod
    def prepare_url(args):
        """ ... """
        pass

    @staticmethod
    def get_image(data):
        """ ... """
        if data:
            if int(data.find('posts')['count']) > 0:
                imagelist = [tag.get('file_url') for tag in data.find_all(
                    'post')]
                return random.choice(imagelist)
        raise NoResultsFound


class BaseSearchHTML:
    """ base html search class. """
    @staticmethod
    def prepare_url(args):
        """ ... """
        pass

    @staticmethod
    def get_image(data):
        """ ... """
        if data:
            images = data.find_all(attrs="thumb")
            if images:
                return random.choice(images).find("img").get("src")
        raise NoResultsFound


class BaseSearchJSON:
    """ base json search class. """
    @staticmethod
    def prepare_url(args):
        """ ... """
        pass

    @staticmethod
    def get_image(data):
        """ ... """
        if data:
            return random.choice(data)['file_url']
        raise NoResultsFound



"""
Read the license at:
https://github.com/IzunaDevs/nsfw_dl/blob/master/LICENSE
"""
__all__ = ['NoLoader', 'NoResultsFound', 'NoXMLParser',
           'UnsupportedDataFormat']


class NoLoader(Exception):
    """
    Thrown when there is no such loader in the NSFWDL instance
    """
    pass


class NoResultsFound(Exception):
    """
    Thrown when the search found no results for the search.
    """
    pass


class NoXMLParser(Exception):
    """
    Thrown when there is no xml parser.
    """
    pass


class UnsupportedDataFormat(Exception):
    """
    Thrown when there is an unsupported format.
    """
    pass



"""
Read the license at:
https://github.com/IzunaDevs/nsfw_dl/blob/master/LICENSE
"""

from .dl import NSFWDL, LOADERS as SOURCES

__title__ = 'nsfw_dl'
__author__ = 'IzunaDevs'
__license__ = 'MIT'
__copyright__ = 'Copyright 2017 IzunaDevs'
__version__ = '0.4.0'
__all__ = ["NSFWDL", "SOURCES"]



"""
Read the license at:
https://github.com/IzunaDevs/nsfw_dl/blob/master/LICENSE
"""
from nsfw_dl.bases import BaseSearchHTML


class GelbooruRandom:
    """ Gets a random image from gelbooru. """
    data_format = "bs4/html"

    @staticmethod
    def prepare_url(args):
        """ .... """
        type(args)
        return "https://www.gelbooru.com/index.php?page=post&s=random", {}, {}

    @staticmethod
    def get_image(data):
        """ .... """
        try:
            return f'https:{data.find(id="image").get("src")}'
        except AttributeError:
            raise AttributeError(str(data))


class GelbooruSearch(BaseSearchHTML):
    """ Gets a random image with a specific tag from gelbooru. """
    data_format = "bs4/html"

    @staticmethod
    def prepare_url(args):
        """ .... """
        return ("https://www.gelbooru.com/index.php"
                f"?page=post&s=list&tags={args}", {}, {})



"""
Read the license at:
https://github.com/IzunaDevs/nsfw_dl/blob/master/LICENSE
"""
from nsfw_dl.bases import BaseSearchJSON


class KonachanRandom:
    """
    Gets a random image from konachan.
    """
    data_format = "bs4/html"

    @staticmethod
    def prepare_url(args):
        """ ... """
        type(args)
        return "https://konachan.com/post/random", {}, {}

    @staticmethod
    def get_image(data):
        """ ... """
        return f'https:{data.find(id="highres").get("href")}'


class KonachanSearch(BaseSearchJSON):
    """ Gets a random image with a specific tag from konachan. """
    data_format = "json"

    @staticmethod
    def prepare_url(args):
        """ ... """
        return ("https://konachan.com/post.json"
                f"?page=dapi&s=post&q=index&tags={args}", {}, {})



"""
Read the license at:
https://github.com/IzunaDevs/nsfw_dl/blob/master/LICENSE
"""


class DrunkenpumkenRandom:
    """ Gets a random image from drunkenpumken. """
    data_format = "bs4/html"

    @staticmethod
    def prepare_url(args):
        """ ... """
        type(args)
        return ("http://drunkenpumken.booru.org/"
                "index.php?page=post&s=random", {}, {})

    @staticmethod
    def get_image(data):
        """ ... """
        return data.find(id="image").get("src")



"""
Read the license at:
https://github.com/IzunaDevs/nsfw_dl/blob/master/LICENSE
"""
import argparse
import sys

import nsfw_dl


def download(downloader, args, file, download_file):
    with nsfw_dl.NSFWDL() as dl:
        img = dl.download(downloader, args=args)

        if callable(file):
            file = file(img)

        if download_file:
            with open(file, "wb") as f:
                f.write(dl.get(img))
                print(file)

        else:
            print(img)


def main(argv):
    """
    Main entrypoint to nsfw_dl commandline.
    """
    parser = argparse.ArgumentParser()
    image = argparse.ArgumentParser()
    parser.add_argument("action", choices=["image", "sources"])

    image.add_argument('-d', '--download',
                       help='Download the result to a file.',
                       default=False, action="store_true")
    image.add_argument('-f', '--file',
                       help="Filename to download to.",
                       default=lambda x: x.split("/")[-1])
    image.add_argument('source', help="Image source to use.")
    image.add_argument('query', help="Tags to use during search.",
                       default='', nargs="*")

    args = parser.parse_args(argv)

    if args.action == "sources":
        sources = "\n".join("\n".join(v for v in source) for source in
                            nsfw_dl.SOURCES)
        print(sources)

    else:
        args = image.parse_args(argv[1:])
        download(args.source, args.query, args.file, args.download)


if __name__ == '__main__':
    main(sys.argv[1:])



"""
Read the license at:
https://github.com/IzunaDevs/nsfw_dl/blob/master/LICENSE
"""
from nsfw_dl.bases import BaseSearchXML


class TbibRandom:
    """ Gets a random image from tbib. """
    data_format = "bs4/html"

    @staticmethod
    def prepare_url(args):
        """ ... """
        type(args)
        return "https://www.tbib.org/index.php?page=post&s=random", {}, {}

    @staticmethod
    def get_image(data):
        """ ... """
        return f'https:{data.find(id="image").get("src")}'


class TbibSearch(BaseSearchXML):
    """ Gets a random image with a specific tag from tbib. """
    data_format = "bs4/xml"

    @staticmethod
    def prepare_url(args):
        """ ... """
        return ("https://www.tbib.org/index.php?page="
                f"dapi&s=post&q=index&tags={args}", {}, {})



"""
Read the license at:
https://github.com/IzunaDevs/nsfw_dl/blob/master/LICENSE
"""


class TsuminoRandom:
    """ Gets a random image from tsumino. """
    data_format = "url"

    @staticmethod
    def prepare_url(args):
        """ ... """
        type(args)
        return "http://www.tsumino.com/Browse/Random", {}, {}

    @staticmethod
    def get_image(data):
        """ ... """
        return data.find(id="highres").get("href")



from nsfw_dl import __version__
from setuptools import setup, find_packages
from setuptools.command.test import test as TestCommand


install_requires = [
    'aiohttp>=2.2.0', 'beautifulsoup4>=4.6.0',
    'lxml>=3.8.0']


try:
    with open('README.rst') as f:
        readme = f.read()
except FileNotFoundError:
    readme = ""


class PyTest(TestCommand):
    user_options = []

    def run(self):
        import subprocess
        import sys
        errno = subprocess.call([sys.executable, '-m', 'pytest', 'tests'])
        raise SystemExit(errno)


# TODO: Change classifiers on this.
setup(name='nsfw_dl',
      author='IzunaDevs',
      author_email='izunadevs@martmists.com',
      url='https://github.com/IzunaDevs/nsfw_dl',
      version=__version__,
      packages=find_packages(),
      license='MIT',
      description='Python 3 package for downloading nsfw images.',
      long_description=readme,
      maintainer_email='izunadevs@martmists.com',
      download_url='https://github.com/IzunaDevs/nsfw_dl',
      include_package_data=True,
      install_requires=install_requires,
      tests_require=['pytest', 'flake8', 'pyflakes', 'coverage',
                     'isort', 'pytest-cov', 'pytest-mock',
                     'pytest-timeout', 'pytest-asyncio',
                     'pylint'],
      platforms='Any',
      classifiers=[
          'Development Status :: 4 - Beta',
          'Intended Audience :: Developers',
          'License :: OSI Approved :: MIT License',
          'Natural Language :: English',
          'Operating System :: OS Independent',
          'Programming Language :: Python :: 3.6',
          'Topic :: Software Development :: Libraries',
          'Topic :: Software Development :: Libraries :: Python Modules',
          'Topic :: Utilities'
      ],
      cmdclass=dict(test=PyTest))



"""
Read the license at:
https://github.com/IzunaDevs/nsfw_dl/blob/master/LICENSE
"""
from nsfw_dl.bases import BaseSearchJSON


class E621Random:
    """ Gets a random image from e621. """
    data_format = "bs4/html"

    @staticmethod
    def prepare_url(args):
        """ ... """
        type(args)
        return "https://e621.net/post/random", {}, {}

    @staticmethod
    def get_image(data):
        """ ... """
        return data.find(id="highres").get("href")


class E621Search(BaseSearchJSON):
    """ Gets a random image with a specific tag from e621. """
    data_format = "json"

    @staticmethod
    def prepare_url(args):
        """ ... """
        return ("https://e621.net/post/index.json?page=dapi&s="
                f"post&q=index&tags={args}", {}, {})
